# Copyright (c) 2011 Cloudera, Inc. All rights reserved.

import logging
import time

from cmf.monitor import schema
from cmf.monitor.abstract_mgmt_monitor import AbstractMgmtMonitor
from cmf.monitor.base import DebugStatusReport
from cmf.monitor.json_metrics_utils import ParseError
from cmf.monitor.role_update import RoleUpdate
from cmf.monitor.service_update import ServiceUpdate
from cmf.monitor.servicemonitor.json_parser import JsonParser

from cmf.monitor.proc_metrics_utils import add_proc_metrics

LOG = logging.getLogger('ServiceMonitorMonitor')

_ROLETYPE = "SERVICEMONITOR"

# Section and key names in cloudera-monitor.properties
# Since 4.0
_SECTION = 'servicemonitor'
_SERVICE_NAME = 'service_name'
_ROLE_NAME = 'role_name'
_HOST = 'host'
_FIREHOSE_DEBUG_PORT = 'firehose_debug_port'
_COLLECT_INTERVAL = 'collect_interval'
# Added in CM 4.1
_LOG_DIR = 'log_dir'

class ServiceMonitorMonitor(AbstractMgmtMonitor):

  def __init__(self, daemon, registry):
    AbstractMgmtMonitor.__init__(self,
                                 daemon,
                                 registry,
                                 _SECTION,
                                 _FIREHOSE_DEBUG_PORT)
    self._last_conf_path = None
    self._metrics_tuple = (None, None, None)
    self._last_reported_timestamp = None
    self._last_collection_ended = None
    self._last_collection_length = None
    self._role_dir_monitors = [ ]

    # Config data
    self._name = None
    self._service_name = None
    self._log_dir = None

  def __str__(self):
    return "ServiceMonitorMonitor for %s" % (self._name,)

  def _reload_config(self, confparse):
    """Overrides AbstractMonitor. Returns True/False for success"""
    try:
      service = confparse.get(_SECTION, _SERVICE_NAME)
      name = confparse.get(_SECTION, _ROLE_NAME)
      log_dir = confparse.get_with_default(_SECTION, _LOG_DIR, None)
      host = confparse.get(_SECTION, _HOST)
      firehose_debug_port = confparse.getint(_SECTION, _FIREHOSE_DEBUG_PORT)
    except ValueError, ex:
      self.record_error(ex)
      return False

    metrics_url = self._compute_metrics_url(host, firehose_debug_port)
    if metrics_url is None:
      return False

    self._service_name = service
    self._name = name
    self._log_dir = log_dir
    self._metrics_url = metrics_url

    # Make a monitor for the log directory.
    self._role_dir_monitors = [ ]
    if self._log_dir is not None:
      self.create_dir_monitor(self._log_dir, self._role_dir_monitors)

    return True

  def get_role_update(self):
    """Returns a RoleUpdate to be sent to the firehose"""
    update = RoleUpdate(name=self._name,
                        roletype=_ROLETYPE,
                        service_name=self._service_name)

    if self.add_metrics_from_webserver(self._metrics_tuple,
                                       self._last_reported_timestamp,
                                       update):
      self._last_reported_timestamp = update.timestamp

    add_proc_metrics(update, self._pid, LOG)

    for mon in self._role_dir_monitors:
      update.add_dir_update(mon.get_role_dir_update())

    return update

  def add_webserver_metrics_to_role_update(self, update, metrics):
    if metrics.event_stats is not None:
      update.add_metric(schema.EVENT_PUBLISHER_SEND_QUEUE_SIZE,
                        metrics.event_stats.SendQueueSize)
      update.add_metric(schema.EVENT_PUBLISHER_DROP_COUNT,
                        metrics.event_stats.DropCount)
      update.add_metric(schema.EVENT_PUBLISHER_EXCEPTION_COUNT,
                        metrics.event_stats.ExceptionCount)
      update.add_metric(schema.EVENT_PUBLISHER_PUSHBACK_COUNT,
                        metrics.event_stats.PushbackCount)
      update.add_metric(schema.EVENT_PUBLISHER_SENT_COUNT,
                        metrics.event_stats.SentCount)

    if metrics.role_stage is not None:
      update.add_metric(schema.SMON_ROLE_DROPS,
                        metrics.role_stage.EventsDropped)
      update.add_metric(schema.SMON_ROLE_PROCESSED,
                        metrics.role_stage.EventsProcessed)
      update.add_metric(schema.SMON_ROLE_QUEUE_SIZE,
                        metrics.role_stage.InputQueueSize)

  def get_debug_status_report(self):
    return DebugStatusReport(config_items=self.conf_parser.items(_SECTION),
                             errors=(self.last_error,),
                             last_collection_ended=self._last_collection_ended,
                             last_collection_length=self._last_collection_length)

  def collect(self):
    """Overrides AbstractMonitor. Collect metrics. Invoked asynchronously."""
    start = time.time()

    # Reload config if necessary
    confparse = self.conf_parser
    if self._last_conf_path != confparse.path:
      if not self._reload_config(confparse):
        return
      self._last_conf_path = confparse.path

    # Collect directory info
    for mon in self._role_dir_monitors:
      mon.collect()

    # If our pid value is None, it means that the associated process is not
    # running. If this is the case, we skip metric collection.
    if self._pid is None:
      return

    LOG.debug("Fetching metrics at '%s'" % (self._metrics_url,))
    self._metrics_tuple = self.collect_metrics_from_url(self._metrics_url)

    self._last_collection_ended = time.time()
    self._last_collection_length = self._last_collection_ended - start

  def parse_metrics(self, json):
    return JsonParser(json)

  def get_subject_type(self):
    return "SERVICEMONITOR"
